/*
 *  Copyright (C) 2006-2010 Leandro Nini
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "sid.h"

#ifdef HAVE_SIDPLAY2

#ifdef HAVE_SIDPLAY_BUILDERS_RESID_H
# include <sidplay/builders/resid.h>
#endif

#ifdef HAVE_SIDPLAY_BUILDERS_HARDSID_H
# include <sidplay/builders/hardsid.h>
#endif

#ifdef HAVE_SIDUTILS
# include <sidplay/utils/SidFilter.h>
#endif

#include <FX88591Codec.h>
#include "utils.h"

// created by reswrap from file sid.gif
extern const unsigned char iconSid[126]={
  0x47,0x49,0x46,0x38,0x39,0x61,0x10,0x00,0x10,0x00,0xf2,0x00,0x00,0x00,0x00,0x00,
  0x00,0x00,0x86,0x00,0x00,0xff,0x86,0x00,0x00,0xff,0x00,0x00,0x86,0x82,0x86,0xcc,
  0xcc,0xcc,0x00,0x00,0x00,0x21,0xf9,0x04,0x01,0x00,0x00,0x06,0x00,0x2c,0x00,0x00,
  0x00,0x00,0x10,0x00,0x10,0x00,0x00,0x03,0x43,0x68,0xba,0xdc,0x6e,0x22,0x3e,0x16,
  0x6b,0x99,0xb0,0xc6,0xfb,0x44,0xf0,0x45,0xa8,0x09,0xcc,0x17,0x84,0x4a,0xa0,0x9e,
  0x9c,0x89,0xa6,0x00,0xbb,0xac,0xdc,0x12,0xd6,0x06,0xcd,0x10,0xbc,0xb3,0x32,0x83,
  0xe0,0xa0,0xa1,0x8a,0x2d,0x06,0x00,0xe4,0x90,0x11,0x6b,0x06,0x0c,0xb7,0xd7,0xc2,
  0x09,0xc0,0x3d,0x0a,0x80,0xac,0x75,0x72,0xc3,0x78,0x17,0x09,0x00,0x3b
  };

//sid|dat
#define EXT	"*.sid"

// HVSC path to STIL.
#define HVSC_STIL	"/DOCUMENTS/STIL.txt"

// HVSC path to BUGlist.
#define HVSC_BUGLIST	"/DOCUMENTS/BUGlist.txt"

#ifndef ALAN_SIDPLAY
#  define CREDITS	"Sidplay2\nCopyright \302\251 Simon White"
#  define LINK		"http://sourceforge.net/projects/sidplay2/"
#else
#  define CREDITS	"Sidplay2\nCopyright \302\251 Simon White\nModified by Antti Lankila"
#  define LINK		"http://sourceforge.net/projects/sidplay-residfp/"
#endif

const char sid::name[]="Sidplay2";

const char sid::_engines[2][8] = {"reSID", "hardSID"};

sidConfig_t sid::_settings;

/*****************************************************************/

size_t sid::fillBuffer(void* buffer, const size_t bufferSize, const unsigned int seconds) {

#ifdef HAVE_SIDUTILS
	if (_length && (seconds>=_length))
		return 0;
#endif

#ifndef ALAN_SIDPLAY
	return _sidplay2->play((void*)buffer, bufferSize);
#else
	return _sidplay2->play((float*)buffer, bufferSize/sizeof(float))*2; // FIXME
#endif
}

/*****************************************************************/

bool sid::supports(const char* fileName) {

#if FOX_MINOR>6
	return FXPath::match(FXPath::name(fileName), EXT, NoEscape|PathName|CaseFold);
#else
	return FXPath::match(FXString(EXT), FXPath::name(fileName), FILEMATCH_FILE_NAME|FILEMATCH_NOESCAPE|FILEMATCH_CASEFOLD);
#endif
}

inline FXString sid::ext() const { return FXString(EXT); }

sid::sid() :
	inputPlugin(name, iconSid),
	_sidplay2(0),
	_stil(0),
#ifdef HAVE_SIDUTILS
	_db(0),
	_length(0),
#endif
	_tune(0)
 {
	loadSettings();

	openHvsc(_settings.hvscPath);
}

sid::~sid() {

	close();

#ifdef HAVE_SIDUTILS
	delete _db;
#endif
	delete _stil;
}

void sid::loadSettings() {

	_settings.samplerate=load("Frequency", 44100);
	_settings.channels=load("Channels", 1);
#ifndef ALAN_SIDPLAY
	_settings.precision=(load("Bits", 16)==16)?S16:U8;
	_settings.optimisation=load("Optimisation", 0);
#endif
	_settings.environment=(sid2_env_t)load("Environment", sid2_envPS);
	_settings.clock=(sid2_clock_t)load("Clock speed", SID2_CLOCK_CORRECT);
	_settings.model=(sid2_model_t)load("SID model", SID2_MODEL_CORRECT);
	_settings.forceClock=load("Force clock", FALSE);
	_settings.engine=load("Engine", _engines[0]);
	_settings.filter=load("Filter", TRUE);
	_settings.hvscPath=load("HVSC", FXString::null);
#ifdef HAVE_SIDUTILS
	_settings.filterIni=load("Filter ini", FXString::null);
#endif
}

void sid::saveSettings() {

	if (compare(_settings.hvscPath, load("HVSC", FXString::null))) {
		FXTRACE((10, "Reloading HVSC from %s\n", _settings.hvscPath.text()));
		openHvsc(_settings.hvscPath);
	}

	save("Frequency", _settings.samplerate);
	save("Channels", _settings.channels);
#ifndef ALAN_SIDPLAY
	save("Bits", (_settings.precision==S16)?16:8);
	save("Optimisation", _settings.optimisation);
#endif
	save("Environment", _settings.environment);
	save("Clock speed", _settings.clock);
	save("SID model", _settings.model);
	save("Force clock", _settings.forceClock);
	save("Engine", _settings.engine);
	save("Filter", _settings.filter);
	save("HVSC", _settings.hvscPath);
#ifdef HAVE_SIDUTILS
	save("Filter ini", _settings.filterIni);
#endif
}

bool sid::open(const FXString& fileName) {

	close();

	_sidplay2=new sidplay2;
	sidbuilder *emuSid=0;
#ifdef HAVE_SIDPLAY_BUILDERS_RESID_H
	if (!comparecase(_settings.engine, _engines[0])) {
		ReSIDBuilder *tmpResid=new ReSIDBuilder("FAP reSID");
		tmpResid->create(_sidplay2->info().maxsids);
#ifndef ALAN_SIDPLAY
		tmpResid->sampling(_settings.samplerate);
#endif
		tmpResid->filter(_settings.filter);
		const sid_filter_t* filter=0;
#ifdef HAVE_SIDUTILS
		SidFilter sidFilter;
		sidFilter.read(_settings.filterIni.text());
		filter=sidFilter.provide();
		if (!sidFilter)
			fxmessage("%s\n", sidFilter.error());
#endif
		tmpResid->filter(filter);
		emuSid=(sidbuilder*)tmpResid;
	}
#endif
#ifdef HAVE_SIDPLAY_BUILDERS_HARDSID_H
	if (!comparecase(_settings.engine, _engines[1])) {
		HardSIDBuilder *tmpHardsid=new HardSIDBuilder("FAP hardSID");
		tmpHardsid->create(_sidplay2->info().maxsids);
		tmpHardsid->filter(_settings.filter);
		emuSid=(sidbuilder*)tmpHardsid;
	}
#endif
	if (!emuSid) {
		delPtr(_sidplay2);
		return false;
	}

	sid2_config_t cfg;
	cfg=_sidplay2->config();
	cfg.clockForced=_settings.forceClock;
	cfg.clockSpeed=_settings.clock;
	cfg.environment=_settings.environment;
	cfg.frequency=_settings.samplerate;
#ifndef ALAN_SIDPLAY
	cfg.optimisation=_settings.optimisation;
	cfg.precision=(_settings.precision==S16)?16:8;
	cfg.sampleFormat=(_settings.precision==S16)?SID2_LITTLE_SIGNED:SID2_LITTLE_UNSIGNED;
#endif
	cfg.playback=(_settings.channels==2)?sid2_stereo:sid2_mono;
	cfg.sidEmulation=emuSid;
	cfg.sidModel=_settings.model;
	cfg.sidSamples=true;
	_sidplay2->config(cfg);

#ifdef HAVE_SIDUTILS
	_tune=new SidTuneMod(fileName.text());
	_tune->createMD5(_md5);
#else
	_tune=new SidTune(fileName.text());
#endif
	loadTune(0);

	const FX88591Codec codec;
	if (_sidplay2->info().tuneInfo->numberOfInfoStrings==3) {
		_metaData.addInfo(metaData::TITLE, codec.mb2utf(_sidplay2->info().tuneInfo->infoString[0]).text());
		_metaData.addInfo(metaData::ARTIST, codec.mb2utf(_sidplay2->info().tuneInfo->infoString[1]).text());
		_metaData.addInfo("Released", codec.mb2utf(_sidplay2->info().tuneInfo->infoString[2]).text());
	}

	if (_stil) {
		FXTRACE((10, "Retrieving STIL info\n"));
		FXString comment=FXString(_stil->getAbsGlobalComment(fileName.text()));
		if (!comment.empty())
			comment.append('\n');

		comment.append(FXString(_stil->getAbsEntry(fileName.text())));

		FXString bug=FXString(_stil->getAbsBug(fileName.text()));
		if (!bug.empty()) {
			comment.append('\n');
			comment.append(bug);
		}
		_metaData.addInfo(metaData::COMMENT, codec.mb2utf(comment.text()).text());
	}

	songLoaded(fileName.text());
	return true;
}

void sid::close() {

	delPtr(_tune);
	if (_sidplay2) {
		const sidbuilder *emuSid=_sidplay2->config().sidEmulation;
		delete emuSid;
		delPtr(_sidplay2);
	}

	songLoaded(0);
}

const bool sid::rewind() {

	if (_sidplay2) {
		_sidplay2->stop();
		return true;
	}
	return false;
}

bool sid::subtune(const unsigned int i) {

	if (_tune && (i<=_tune->getInfo().songs)) {
		loadTune(i);
		return true;
	}
	return false;
}

void sid::loadTune(const int num) {

	_tune->selectSong(num);
	_sidplay2->load(_tune);
#ifdef HAVE_SIDUTILS
	_length=_db?_db->length(_md5, _tune->getInfo().currentSong):0;
	time(_length);
#endif
}

void sid::openHvsc(const FXString& hvscPath) {

	if (hvscPath.empty())
		return;

	if (!_stil)
		_stil=new STIL(HVSC_STIL, HVSC_BUGLIST);

	if (!_stil->setBaseDir(hvscPath.text())) {
		fxmessage("%s\n", _stil->getErrorStr());
		delPtr(_stil);
	}

#ifdef HAVE_SIDUTILS
	if (!_db)
		_db=new SidDatabase();

	if (_db->open(FXStringFormat("%s%cDOCUMENTS%cSonglengths.txt", hvscPath.text(), PATHSEP, PATHSEP).text())) {
		fxmessage("%s\n", _db->error());
		delPtr(_db);
	}
#endif
}

/*****************************************************************/

#define SETTINGS	sid::_settings

FXDEFMAP(sidConfig) sidConfigMap[] = {
	// Message_Type		ID				Message_Handler
	FXMAPFUNC(SEL_COMMAND,	sidConfig::ID_FREQUENCY,	sidConfig::onCmdFrequency),
	FXMAPFUNC(SEL_COMMAND,	sidConfig::ID_CHANNELS,		sidConfig::onCmdChannels),
#ifndef ALAN_SIDPLAY
	FXMAPFUNC(SEL_COMMAND,	sidConfig::ID_BITS,		sidConfig::onCmdBits),
#endif
	FXMAPFUNC(SEL_COMMAND,	sidConfig::ID_ENGINE,		sidConfig::onCmdEngine),
	FXMAPFUNC(SEL_COMMAND,	sidConfig::ID_ENVIRONMENT,	sidConfig::onCmdEnvironment),
	FXMAPFUNC(SEL_COMMAND,	sidConfig::ID_HVSC,		sidConfig::onCmdHvsc),
	FXMAPFUNC(SEL_COMMAND,	sidConfig::ID_FILTER,		sidConfig::onCmdFilter),
};

FXIMPLEMENT(sidConfig, configFrame, sidConfigMap, ARRAYNUMBER(sidConfigMap))

sidConfig::sidConfig(FXPacker* win) :
	configFrame(win, sid::name, CREDITS, LINK) {

	FXint val;

#ifndef ALAN_SIDPLAY
	_optimisation.connect(SETTINGS.optimisation);
#endif
	_filter.connect(SETTINGS.filter);
	_hvscDir.connect(SETTINGS.hvscPath);
#ifdef HAVE_SIDUTILS
	_filterIni.connect(SETTINGS.filterIni);
#endif

	new FXLabel(matrix(), tr("Frequency"), 0, LAYOUT_FILL|JUSTIFY_RIGHT);
	FXListBox *freqBox=new FXListBox(matrix(), this, sidConfig::ID_FREQUENCY, LAYOUT_FILL|FRAME_NORMAL);
	freqBox->fillItems("11025\n22050\n44100\n48000");
	freqBox->setNumVisible(4);
	switch (SETTINGS.samplerate) {
	case 11025:
		val=0;
		break;
	case 22050:
		val=1;
		break;
	default:
	case 44100:
		val=2;
		break;
	case 48000:
		val=3;
		break;
	}
	freqBox->setCurrentItem(val);

	new FXLabel(matrix(), tr("Channels"), 0, LAYOUT_FILL|JUSTIFY_RIGHT);
	FXListBox *chanBox=new FXListBox(matrix(), this, sidConfig::ID_CHANNELS, LAYOUT_FILL|FRAME_NORMAL);
	chanBox->fillItems("Mono\nStereo");
	chanBox->setNumVisible(2);
	chanBox->setCurrentItem(SETTINGS.channels-1);
#ifndef ALAN_SIDPLAY
	new FXLabel(matrix(), tr("Bits"), 0, LAYOUT_FILL|JUSTIFY_RIGHT);
	FXListBox *_bitBox=new FXListBox(matrix(), this, sidConfig::ID_BITS, LAYOUT_FILL|FRAME_NORMAL);
	_bitBox->fillItems("8\n16");
	_bitBox->setNumVisible(2);
	switch (SETTINGS.precision) {
	case U8:
		val=0;
		break;
	default:
	case S16:
		val=1;
		break;
	}
	_bitBox->setCurrentItem(val);
#endif
	new FXLabel(matrix(), tr("Environment"), 0, LAYOUT_FILL|JUSTIFY_RIGHT);
	FXListBox *resBox=new FXListBox(matrix(), this, sidConfig::ID_ENVIRONMENT, LAYOUT_FILL|FRAME_NORMAL);
	resBox->fillItems("Playsid\nTransparent Rom\nBankswitching\nReal C64 Environment");
	resBox->setNumVisible(4);
	switch (SETTINGS.environment) {
	default:
	case sid2_envPS:
		val=0;
		break;
	case sid2_envTP:
		val=1;
		break;
	case sid2_envBS:
		val=2;
		break;
	case sid2_envR:
		val=3;
		break;
	case sid2_envTR:
		val=4;
		break;
	}
	resBox->setCurrentItem(val);

	new FXLabel(matrix(), tr("Engine"), 0, LAYOUT_FILL|JUSTIFY_RIGHT);
	FXListBox *engBox=new FXListBox(matrix(), this, sidConfig::ID_ENGINE, LAYOUT_FILL|FRAME_NORMAL);
#ifdef HAVE_SIDPLAY_BUILDERS_RESID_H
	engBox->appendItem(sid::_engines[0]);
#endif

#ifdef HAVE_SIDPLAY_BUILDERS_HARDSID_H
	engBox->appendItem(sid::_engines[1]);
#endif
	const FXint numItems=engBox->getNumItems();
	engBox->setNumVisible(numItems);
	const FXint curItem=engBox->findItem(SETTINGS.engine);
	if (curItem>=0)
		engBox->setCurrentItem(curItem);

	new FXVerticalSeparator(extraLeft());

	FXVerticalFrame *vert=new FXVerticalFrame(extraLeft());
#ifndef ALAN_SIDPLAY
	new FXCheckButton(vert, tr("Optimisation\tFaster but inaccurate emulation"), &_optimisation, FXDataTarget::ID_VALUE);
#endif
	new FXCheckButton(vert, tr("Filter\tEmulate SID filter"), &_filter, FXDataTarget::ID_VALUE);

	FXMatrix *frame=new FXMatrix(extraBottom(), 3, MATRIX_BY_COLUMNS);

	new FXLabel(frame, tr("HVSC path:"));
	new FXTextField(frame, 40, &_hvscDir, FXDataTarget::ID_VALUE, TEXTFIELD_READONLY|TEXTFIELD_NORMAL);
	new FXButton(frame, tr("&Browse\tSelect HVSC directory"), 0, this, sidConfig::ID_HVSC, FRAME_RAISED|FRAME_THICK|JUSTIFY_LEFT|ICON_BEFORE_TEXT);

#ifdef HAVE_SIDUTILS
	new FXLabel(frame, tr("Filter ini:"));
	new FXTextField(frame, 40, &_filterIni, FXDataTarget::ID_VALUE, TEXTFIELD_READONLY|TEXTFIELD_NORMAL);
	new FXButton(frame, tr("&Browse\tSelect HVSC directory"), 0, this, sidConfig::ID_FILTER, FRAME_RAISED|FRAME_THICK|JUSTIFY_LEFT|ICON_BEFORE_TEXT);
#endif
}

long sidConfig::onCmdFrequency(FXObject*, FXSelector, void* ptr) {

	const FXint val=(FXint)(FXival)ptr;
	switch (val) {
	case 0:
		SETTINGS.samplerate=11025;
		break;
	case 1:
		SETTINGS.samplerate=22050;
		break;
	case 2:
		SETTINGS.samplerate=44100;
		break;
	case 3:
		SETTINGS.samplerate=48000;
		break;
	}
	return 1;
}

long sidConfig::onCmdChannels(FXObject*, FXSelector, void* ptr) {

	const FXint val=(FXint)(FXival)ptr;
	SETTINGS.channels=val+1;
	return 1;
}
#ifndef ALAN_SIDPLAY
long sidConfig::onCmdBits(FXObject*, FXSelector, void* ptr) {

	const FXint val=(FXint)(FXival)ptr;
	switch (val) {
	case 0:
		SETTINGS.precision=U8;
		break;
	case 1:
		SETTINGS.precision=S16;
		break;
	}
	return 1;
}
#endif
long sidConfig::onCmdEngine(FXObject*, FXSelector, void* ptr) {

	const FXint val=(FXint)(FXival)ptr;
	SETTINGS.engine=sid::_engines[val];
	return 1;
}

long sidConfig::onCmdEnvironment(FXObject*, FXSelector, void* ptr) {

	const FXint val=(FXint)(FXival)ptr;
	switch (val) {
	case 0:
		SETTINGS.environment=sid2_envPS;
		break;
	case 1:
		SETTINGS.environment=sid2_envTP;
		break;
	case 2:
		SETTINGS.environment=sid2_envBS;
		break;
	case 3:
		SETTINGS.environment=sid2_envR;
		break;
	}
	return 1;
}

long sidConfig::onCmdHvsc(FXObject*, FXSelector, void*) {

	FXDirDialog dirDialog(this, tr("Select HVSC directory"));
	dirDialog.setDirectory(SETTINGS.hvscPath);
	if (dirDialog.execute())
		SETTINGS.hvscPath=dirDialog.getDirectory();

	return 1;
}

long sidConfig::onCmdFilter(FXObject*, FXSelector, void*) {

	FXFileDialog fileDialog(this, tr("Select filter ini file"));
	fileDialog.setPattern("*.ini");
	fileDialog.setFilename(SETTINGS.filterIni);
	if (fileDialog.execute())
		SETTINGS.filterIni=fileDialog.getFilename();

	return 1;
}

#endif
